@@ -6,23 +6,21 @@
  * 
  * Copyright (c) 2004 James Su <suzhe@tsinghua.org.cn>
  *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
+ * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU Lesser General Public License for more details.
+ * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
- * Boston, MA  02111-1307  USA
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
- * $Id: scim_m17n_imengine.cpp,v 1.13 2005/04/04 10:17:00 suzhe Exp $
+ * $Id: scim_m17n_imengine.cpp,v 1.25 2006/12/01 08:45:12 suzhe Exp $
  */
 
 #define Uses_SCIM_UTILITY
@@ -43,151 +41,16 @@
 #define scim_imengine_module_init m17n_LTX_scim_imengine_module_init
 #define scim_imengine_module_create_factory m17n_LTX_scim_imengine_module_create_factory
 
-#define SCIM_CONFIG_IMENGINE_M17N_UUID    "/IMEngine/M17N/UUID-"
-
 #define SCIM_PROP_STATUS                  "/IMEngine/M17N/Status"
 
 #ifndef SCIM_M17N_ICON_FILE
     #define SCIM_M17N_ICON_FILE           (SCIM_ICONDIR "/scim-m17n.png")
 #endif
 
-static const size_t  __m17n_nr_uuids = 128;
-static const char   *__m17n_uuids [] = {
-"de94c8e7-6021-4787-a7ed-0d505bf9eea3",
-"31cfe859-8015-4067-9b75-1615bd54c72d",
-"a3e7a2f8-2f39-4586-b851-80a857b1557d",
-"adfbf67c-156d-44bd-95a7-5827286d6500",
-"682a7366-3f4b-4198-8c42-aebc9def0c36",
-"68243143-85b7-4d88-8aeb-b5232c34004a",
-"a5e616db-63eb-4e10-a15d-4811655af4f3",
-"98ec723d-97a3-4cef-9c0a-df9aa62b6e2e",
-"f42fbcda-d2e9-459b-82b2-396b187dd5c5",
-"7a429b16-8527-4e9e-9783-b30b7d17adaf",
-"d5a0cf07-3b36-4c68-b02d-7ef2c78d8a75",
-"22879d8e-4824-4f70-947c-c0db8037ecd6",
-"65b2743c-c1f0-48c1-be50-ce69a7fa1949",
-"8969bed5-eb7b-441c-8d5e-1de5cf463328",
-"2523be1e-b11c-4921-a1ec-67e483650039",
-"e27d65e3-db40-42ef-a81b-57b762ff7cc0",
-"c5af66e1-f017-490c-9598-fe68fe70818f",
-"12dc8499-4e99-4237-96f6-555a6d6026f0",
-"0f730558-23d2-493a-8bc7-0473ff6125c7",
-"3a80941c-2524-4847-9ded-bf44a355f4de",
-"3ad143b0-480d-48ed-bfb4-9b9b9ba71c0c",
-"415184fb-d645-496f-bbde-f4f8c253395d",
-"ae27032d-6c31-4018-9d13-f1839aab6545",
-"3d102755-fc08-4e24-9b74-218d044f4a29",
-"7dd0fa4a-93ce-466b-8dad-aec41a2a549e",
-"c7b0e37d-4ed0-41a9-8e55-6a882889b09c",
-"f32952ad-4279-4999-8895-ebe6d6c8aa84",
-"b7e3d4e6-e3e0-4a5b-accd-7f911cbdcd36",
-"2bd9599d-708a-4af3-957c-d2621bc622a4",
-"5b5edc0f-510a-49e0-bcdc-cdb2074ae682",
-"d4c08360-d8de-4551-9c9a-e28cb4636a0b",
-"338ecf26-9e73-4dda-ac73-29109fdfdc5e",
-"992b8610-0173-4bdc-bd0b-5e9016219d33",
-"29e26490-02c9-4d43-9668-c043c2e5e3e7",
-"f1ccfd6b-88ba-48e8-a359-7df7371bcd82",
-"1b76eb2c-7982-47d3-af9c-005577e11b12",
-"047a41c8-5da0-4aca-a35f-5f23078840f1",
-"0fc676c9-943a-43f4-be05-e7e43b13b96a",
-"3da743d7-4d5b-4fe4-aa46-597d176fb452",
-"775759d7-49f9-4dd0-a9b6-8ae81bcd899c",
-"234f1592-3180-480c-acf3-2c68e8a4a302",
-"339f0bfe-b947-4f09-b0b0-2682e34872e3",
-"c68ad9a6-8404-4575-9aba-67de51fe6d60",
-"4d6e74ac-d752-4173-a732-a1f667cb1526",
-"b4c8aca2-7cd3-4d3a-a12a-6797fed89851",
-"963a15d2-2384-4543-ac43-e3d1e03c0305",
-"6fd6a47a-2178-437f-9f9d-1ef70f4bb7d7",
-"4b122e93-b6c1-4625-ae4b-1461ee12ade3",
-"439ee607-2e5d-44eb-a164-a6a3122bd5b9",
-"d15a79a9-1f21-438c-9d25-22dca6a3ad38",
-"c19fc0fa-2cfd-449e-82d9-29b07da79b8a",
-"df4e3287-581b-4ce8-9e6a-baebebb0b1a3",
-"84457221-9757-46d4-87c9-27a39b7542cc",
-"9900e41f-9295-45bc-9cb0-440e4198ad46",
-"dea96ada-ea99-4e8a-a0cc-d1b6ae56d702",
-"fe9856d3-1776-47b8-b393-31e2a84431ad",
-"46260747-acd2-4476-864a-8e208e1a48fb",
-"16cefe32-1789-4b1d-b3a5-3cc1e15c8128",
-"dd95ff09-82da-4a3d-b51d-aaeed259002b",
-"d164d5d9-87cb-4ac7-9953-ab28ce37c59f",
-"f74c872c-6ebe-4702-8ef7-c5fd271d36cf",
-"06155ca5-554b-45a7-a04a-907ea38c2979",
-"0a6c1b0c-0c9e-400b-81d0-69b0323fa655",
-"c79aed67-6d75-40af-b2d0-93a1cec5c596",
-"8f0c5c5b-9121-4962-9475-96767eae586e",
-"766656eb-6575-4e16-9d92-c382c1f5f5a2",
-"ff41d2ed-cb47-49b0-920e-06abd0ee277c",
-"7b49235d-7550-4cef-b1d1-0a62fad4bd2c",
-"34cecab5-1f22-431f-a895-e855cbd41bf0",
-"2adfb2f1-6242-4730-996a-18447081670c",
-"3d08446a-389c-4c98-aa84-13960d5cfd3e",
-"92fef7d9-14e1-44df-b5c5-8bd3e4d364af",
-"99bf476f-7746-4b9c-9940-d6f4ca1fcef5",
-"fc738e82-98fc-47a6-87d0-9de582e9f625",
-"8f4b8524-93f6-4792-ac03-f9ca5f0e49ef",
-"c3848b56-0e3c-454c-b859-dddb52a22c42",
-"f5117076-707e-41d5-aef3-381027b020bb",
-"e1dddc6b-c268-4dd5-8629-1d1bdff8efe2",
-"329e9e57-780f-4b1e-8c96-e206cae70fc6",
-"3ec7a83a-883c-4c6d-9f34-f616d7f66227",
-"9b0989dc-047d-4186-93d7-59c4a67792d3",
-"c4366de0-64bd-4672-8613-bf89c269e125",
-"5f11f45c-6545-4d3b-ad95-dc02a1827e55",
-"52ed1916-f753-4eca-89a4-3dd527f6e7dd",
-"5963628b-6cb6-4de5-a531-b05529c3b02f",
-"ed49a819-086f-4e48-b60a-815862049f87",
-"60a4641f-e585-4940-8142-c853c7c81e5a",
-"f85db69b-98b3-42bc-a8f7-c9aff51c93a4",
-"a8866a89-1a9d-4584-a562-3731bb678dae",
-"ab48f3fb-ec6e-4d16-bb89-845122b75f05",
-"5f3881f2-7313-48fa-8ed5-1f3714232829",
-"29cbbd59-ac94-461b-8fdd-537c28c50847",
-"68ddbf8d-ea8a-4163-96ba-e3f711385507",
-"513c0ae8-9a17-4c29-8241-75d803055058",
-"0a87a377-f361-40c7-b183-b74cf378d642",
-"9a418ce6-0ee6-459d-9f2e-69c1d33a854c",
-"d5f5fbcc-2496-41f7-aa5f-d934e88c72fc",
-"e439ccdc-d2a2-4949-b658-691be51866fc",
-"b5e808c0-72f7-4ba7-a4d7-29d8ee6f33b7",
-"022b8e62-5721-4144-8c90-0f87a194c3cb",
-"8db6baad-72cb-485f-859d-7a85d3e2ecc7",
-"1929ea7b-bdf3-4e63-8c81-ebe95e1274a1",
-"f25a4891-1a69-4190-8143-ba2e9d6dfa4e",
-"92a12eba-988c-47d4-a473-8a83814a142e",
-"38c502a8-8de1-47de-a9c1-f1f9a35f4689",
-"740218f0-8aca-433a-88c2-3b2e077df8e4",
-"98d3e84e-e78e-40e4-bf4c-bc95bfcb37a5",
-"b0bd43a3-c537-4934-802b-a4cccd20b7e7",
-"61d4b4ef-b1a1-42fd-9e00-da0e148501f0",
-"b29c4400-b880-4a03-9a80-a1d48dcb8b30",
-"4d00c0b3-4f1e-4b9e-832d-c5904d48e5d9",
-"a1e45eb5-fc35-4455-ab8a-71acbc33d95f",
-"e9c7013a-7630-4310-a4a1-c2ec67f76c7a",
-"04ab7e89-bdba-4f5e-9442-6f653acf04f2",
-"dee60281-0dbb-4d1b-a989-6ee89f4c9250",
-"c0ee3079-2c0c-4b4b-894f-beb5404162f5",
-"fb17e4fc-9901-4c44-9e7f-7b179276066d",
-"d502438f-4e0e-4d39-ab9b-883cf15695c2",
-"f3537cf5-6e82-4162-9eca-ffd1a9bd5bb9",
-"aa91350b-a0b7-4ce2-ac82-1c208dddcfce",
-"2d2d2a6f-64a6-49b4-af11-25c49857bbc6",
-"8ac094fb-2b1b-4eeb-be91-4c87f1c7e34b",
-"7ab49ede-33da-4ce2-9d6c-7ee83a61cc43",
-"1cab3d11-1fe8-4bbe-8759-4844939f486c",
-"f36a0c58-f71e-4539-a43e-cad371c71d52",
-"f6b7b374-be88-4022-ac10-d5c47998a9b4",
-"b6fd6dd7-f4fe-4860-b92d-272f8cba616c",
-"d1c01bd6-ad12-42b8-9b73-47e51fcff098"
-};
-
 struct M17NInfo {
-    String        name;
     String        lang;
+    String        name;
     String        uuid;
-    MInputMethod *im;
 };
 
 static std::vector <M17NInfo>                     __m17n_input_methods;
@@ -195,9 +58,7 @@ static std::map <MInputContext *, M17NInstance *> __m17n_input_contexts;
 
 static CommonLookupTable                          __lookup_table;
 
-static ConfigPointer __scim_config (0);
-
-static MConverter *__m17n_converter = 0;
+static MConverter   *__m17n_converter = 0;
 
 static M17NInstance * __find_instance (MInputContext *ic);
 
@@ -206,19 +67,11 @@ static MSymbol __key_to_symbol (const KeyEvent &key);
 extern "C" {
     void scim_module_init (void)
     {
-        std::vector <WideString> labels;
-
-        for (size_t i = 0; i < 16; ++i)
-            labels.push_back (WideString ());
-
-        __lookup_table.set_candidate_labels (labels);
+        __lookup_table.fix_page_size ();
     }
 
     void scim_module_exit (void)
     {
-        for (size_t i = 0; i < __m17n_input_methods.size (); ++i)
-            if (__m17n_input_methods [i].im)
-                minput_close_im (__m17n_input_methods [i].im);
         M17N_FINI();
     }
 
@@ -227,13 +80,13 @@ extern "C" {
         SCIM_DEBUG_IMENGINE(1) << "Initialize M17N Engine.\n";
 
         MPlist *imlist, *elm;
-        MSymbol utf8 = msymbol("utf8");
+        MSymbol utf8;
+
         size_t i;
         size_t count = 0;
 
-        __scim_config = config;
-
         M17N_INIT();
+        utf8 = msymbol("utf8");
 
         __m17n_converter = mconv_buffer_converter(utf8, NULL, 0);
 
@@ -244,61 +97,33 @@ extern "C" {
         for (elm = imlist; elm && mplist_key(elm) != Mnil; elm = mplist_next(elm)) {
             MDatabase *mdb = (MDatabase *) mplist_value(elm);
             MSymbol *tag = mdatabase_tag(mdb);
-            if (tag[1] != Mnil) {
-                MInputMethod *im = minput_open_im(tag[1], tag[2], NULL);
-                if (im) {
-                    const char *im_lang = msymbol_name (im->language);
-                    const char *im_name = msymbol_name (im->name);
+            if (tag[1] != Mnil && tag[2] != Mnil) {
+                const char *im_lang = msymbol_name (tag[1]);
+                const char *im_name = msymbol_name (tag[2]);
 
-                    if (im_lang && strlen (im_lang) && im_name && strlen (im_name)) {
-                        M17NInfo info;
+                if (im_lang && strlen (im_lang) && im_name && strlen (im_name)) {
+                    M17NInfo info;
 
-                        SCIM_DEBUG_IMENGINE(1) << im_lang << "-" << im_name << "\n";
+                    SCIM_DEBUG_IMENGINE(1) << im_lang << "-" << im_name << "\n";
 
-                        info.name = String (im_lang) + String ("-") + String (im_name);
-                        info.lang = String (im_lang);
-                        info.im = im;
+                    info.lang = String (im_lang);
+                    info.name = String (im_name);
 
-                        __m17n_input_methods.push_back (info);
+                    __m17n_input_methods.push_back (info);
 
-	                    im->driver.callback_list = M17NInstance::register_callbacks(im->driver.callback_list);
-
-                        if (++count >= __m17n_nr_uuids) break;
-
-                    } else {
-                        minput_close_im (im);
-                    }
+                    count++;
                 }
             }
         }
 
-        m17n_object_unref(imlist);
+        if (imlist) m17n_object_unref(imlist);
 
-        // Get all uuids.
+        // Set uuids.
         for (i = 0; i < count; ++i)
-            __m17n_input_methods [i].uuid = config->read (String (SCIM_CONFIG_IMENGINE_M17N_UUID) + __m17n_input_methods [i].name,
-                                                          String (""));
-
-        // Allocate uuids for the IMs without uuid.
-        for (i = 0; i < count; ++i) {
-            if (!__m17n_input_methods [i].uuid.length ()) {
-                for (size_t j = 0; j < __m17n_nr_uuids; ++j) {
-                    size_t k = 0;
-                    for (; k < count; ++k)
-                        if (String (__m17n_uuids [j]) == __m17n_input_methods [k].uuid) break;
-
-                    if (k == count) {
-
-                        SCIM_DEBUG_IMENGINE(1) << "Set UUID: " << __m17n_uuids [j] << " -> " << __m17n_input_methods [i].name << "\n";
-
-                        __m17n_input_methods [i].uuid = __m17n_uuids [j];
-                        config->write (String (SCIM_CONFIG_IMENGINE_M17N_UUID) + __m17n_input_methods [i].name,
-                                       String (__m17n_uuids [j]));
-                        break;
-                    }
-                }
-            }
-        }
+            __m17n_input_methods [i].uuid = String ("IMEngine-M17N-" +
+                                                    __m17n_input_methods [i].lang +
+                                                    String ("-") +
+                                                    __m17n_input_methods [i].name);
 
         return count;
     }
@@ -310,9 +135,8 @@ extern "C" {
         M17NFactory *factory = 0;
 
         try {
-            factory = new M17NFactory (__m17n_input_methods [engine].im,
+            factory = new M17NFactory (__m17n_input_methods [engine].lang,
                                        __m17n_input_methods [engine].name,
-                                       __m17n_input_methods [engine].lang,
                                        __m17n_input_methods [engine].uuid);
         } catch (...) {
             delete factory;
@@ -323,17 +147,14 @@ extern "C" {
     }
 }
 
-M17NFactory::M17NFactory (MInputMethod *im,
+M17NFactory::M17NFactory (const String &lang,
                           const String &name,
-                          const String &lang,
                           const String &uuid)
-    : m_im (im),
-      m_name (name),
-      m_uuid (uuid)
+    : m_im (0), m_lang (lang), m_name (name), m_uuid (uuid)
 {
     SCIM_DEBUG_IMENGINE(1) << "Create M17N Factory :\n";
-    SCIM_DEBUG_IMENGINE(1) << "  Name : " << name << "\n";
     SCIM_DEBUG_IMENGINE(1) << "  Lang : " << lang << "\n";
+    SCIM_DEBUG_IMENGINE(1) << "  Name : " << name << "\n";
     SCIM_DEBUG_IMENGINE(1) << "  UUID : " << uuid << "\n";
 
     if (lang.length () >= 2)
@@ -342,12 +163,31 @@ M17NFactory::M17NFactory (MInputMethod *im,
 
 M17NFactory::~M17NFactory ()
 {
+    if (m_im)
+        minput_close_im (m_im);
+}
+
+bool
+M17NFactory::load_input_method ()
+{
+    SCIM_DEBUG_IMENGINE(1) << "load_input_method(" << m_lang << "," << m_name << ")\n";
+
+    if (m_im) return true;
+
+    m_im = minput_open_im(msymbol (m_lang.c_str ()), msymbol (m_name.c_str ()), NULL);
+
+    if (m_im) {
+        m_im->driver.callback_list = M17NInstance::register_callbacks(m_im->driver.callback_list);
+        return true;
+    }
+
+    return false;
 }
 
 WideString
 M17NFactory::get_name () const
 {
-    return utf8_mbstowcs (String ("M17N-") + m_name);
+    return utf8_mbstowcs (m_lang + String ("-") + m_name);
 }
 
 WideString
@@ -365,6 +205,19 @@ M17NFactory::get_credits () const
 WideString
 M17NFactory::get_help () const
 {
+#if M17N_VERSION >= 10300
+    MText *desc = minput_get_description (msymbol (m_lang.c_str ()), msymbol (m_name.c_str ()));
+    if (desc) {
+        int bufsize = mtext_len (desc) * 6;  // long enough
+        char *buf = new char [ bufsize ];
+        mconv_rebind_buffer(__m17n_converter, (unsigned char *) buf, bufsize);
+        mconv_encode(__m17n_converter, desc);
+        buf[__m17n_converter->nbytes] = 0;
+        m17n_object_unref(desc);
+
+        return utf8_mbstowcs (buf);
+    }
+#endif
     return WideString ();
 }
 
@@ -377,25 +230,53 @@ M17NFactory::get_uuid () const
 String
 M17NFactory::get_icon_file () const
 {
+#if M17N_VERSION >= 10300
+    MPlist *l = minput_get_title_icon (msymbol (m_lang.c_str ()), msymbol (m_name.c_str ()));
+    if (l) {
+        char buf [256] = SCIM_M17N_ICON_FILE;
+        MPlist *n = mplist_next (l);
+
+        if (n && mplist_key (n) == Mtext) {
+            MText *icon = (MText*) mplist_value (n);
+            mconv_rebind_buffer(__m17n_converter, (unsigned char *) buf, 256);
+            mconv_encode(__m17n_converter, icon);
+            buf[__m17n_converter->nbytes] = 0;
+        }
+
+        m17n_object_unref (l);
+        return String (buf);
+    }
+#endif
     return String (SCIM_M17N_ICON_FILE);
 }
 
 IMEngineInstancePointer
 M17NFactory::create_instance (const String &encoding, int id)
 {
-    return new M17NInstance (this, encoding, id);
+    if (m_im || load_input_method ())
+        return new M17NInstance (this, encoding, id);
+
+    return new DummyIMEngineInstance(dynamic_cast<DummyIMEngineFactory*>(this), encoding, id);
 }
 
 M17NInstance::M17NInstance (M17NFactory  *factory,
                             const String &encoding,
                             int           id)
     : IMEngineInstanceBase (factory, encoding, id),
-      m_ic (0)
+      m_ic (0),
+      m_cap (0),
+      m_block_preedit_op (false),
+      m_pending_preedit_start (false),
+      m_pending_preedit_draw (false),
+      m_pending_preedit_done (false),
+      m_preedit_showed (false)
 {
-    SCIM_DEBUG_IMENGINE(1) << "Create M17N Instance.\n";
+    SCIM_DEBUG_IMENGINE(1) << "Create M17N Instance " << encoding << " " << id << "\n";
 
-    if (factory->m_im)
+    if (factory->m_im) {
+        SCIM_DEBUG_IMENGINE(2) << " Create minput instance.\n";
         m_ic = minput_create_ic (factory->m_im, NULL);
+    }
 
     if (m_ic)
         __m17n_input_contexts [m_ic] = this;
@@ -403,6 +284,8 @@ M17NInstance::M17NInstance (M17NFactory  *factory,
 
 M17NInstance::~M17NInstance ()
 {
+    SCIM_DEBUG_IMENGINE(1) << "Destroy M17N Instance " << get_id () << "\n";
+
     if (m_ic) {
         __m17n_input_contexts.erase (m_ic);
         minput_destroy_ic (m_ic);
@@ -416,26 +299,37 @@ M17NInstance::process_key_event (const KeyEvent& key)
 
     if (key.is_key_release ()) return true;
 
-    MSymbol m17n_key = __key_to_symbol (key);
+    MSymbol m17n_key = __key_to_symbol (key.map_to_layout (SCIM_KEYBOARD_Default));
 
     if (m17n_key == Mnil) return false;
 
-    SCIM_DEBUG_IMENGINE(2) << "process_key_event. " << msymbol_name (m17n_key) << "\n";
+    return m17n_process_key (m17n_key);
+}
+
+bool
+M17NInstance::m17n_process_key (MSymbol key)
+{
+    SCIM_DEBUG_IMENGINE(2) << "process_key_event. " << msymbol_name (key) << "\n";
 
     char buf [1024];
     MText *produced;
     int ret;
 
-    ret = minput_filter(m_ic, m17n_key, NULL);
+    m_block_preedit_op = true;
+
+    ret = minput_filter(m_ic, key, NULL);
+
+    m_block_preedit_op = false;
 
     if (ret) {
         SCIM_DEBUG_IMENGINE(3) << "minput_filter returns 1\n";
+        do_preedit_op ();
         return true;
     }
 
     produced = mtext();
 
-    ret = minput_lookup(m_ic, m17n_key, NULL, produced);
+    ret = minput_lookup(m_ic, key, NULL, produced);
 
     if (ret) {
         SCIM_DEBUG_IMENGINE(3) << "minput_lookup returns 1\n";
@@ -451,9 +345,33 @@ M17NInstance::process_key_event (const KeyEvent& key)
         commit_string (utf8_mbstowcs (buf));
     }
 
+    do_preedit_op ();
+
     return ret == 0;
 }
 
+void
+M17NInstance::do_preedit_op ()
+{
+    if (m_block_preedit_op)
+        return;
+
+    if (m_pending_preedit_start) {
+        preedit_start_cb (m_ic, Minput_preedit_start);
+        m_pending_preedit_start = false;
+    }
+
+    if (m_pending_preedit_draw) {
+        preedit_draw_cb (m_ic, Minput_preedit_draw);
+        m_pending_preedit_draw = false;
+    }
+
+    if (m_pending_preedit_done) {
+        preedit_done_cb (m_ic, Minput_preedit_done);
+        m_pending_preedit_done = false;
+    }
+}
+
 void
 M17NInstance::move_preedit_caret (unsigned int pos)
 {
@@ -462,6 +380,11 @@ M17NInstance::move_preedit_caret (unsigned int pos)
 void
 M17NInstance::select_candidate (unsigned int item)
 {
+    if (item <= 10) {
+        char buf [4];
+        snprintf (buf, 4, "%d", (item + 1) % 10);
+        m17n_process_key (msymbol (buf));
+    }
 }
 
 void
@@ -472,16 +395,29 @@ M17NInstance::update_lookup_table_page_size (unsigned int page_size)
 void
 M17NInstance::lookup_table_page_up ()
 {
+    m17n_process_key (msymbol ("Up"));
 }
 
 void
 M17NInstance::lookup_table_page_down ()
 {
+    m17n_process_key (msymbol ("Down"));
 }
 
 void
 M17NInstance::reset ()
 {
+    SCIM_DEBUG_IMENGINE(2) << "reset.\n";
+    minput_reset_ic(m_ic);
+
+    hide_preedit_string ();
+    hide_aux_string ();
+    hide_lookup_table ();
+
+    m_preedit_showed = false;
+    m_pending_preedit_start = false;
+    m_pending_preedit_draw = false;
+    m_pending_preedit_done = false;
 }
 
 void
@@ -489,10 +425,6 @@ M17NInstance::focus_in ()
 {
     SCIM_DEBUG_IMENGINE(2) << "focus_in.\n";
 
-    hide_preedit_string ();
-    hide_aux_string ();
-    hide_lookup_table ();
-
     PropertyList props;
     Property     prop (String (SCIM_PROP_STATUS),
                        String (""));
@@ -502,36 +434,27 @@ M17NInstance::focus_in ()
 
     register_properties (props);
 
+#if M17N_VERSION >= 10300
+    m17n_process_key (Minput_focus_in);
+#else
     preedit_draw_cb (m_ic, Minput_preedit_draw);
-    status_draw_cb (m_ic, Minput_status_draw);
     candidates_draw_cb (m_ic, Minput_candidates_draw);
+#endif
+
+    //FIXME: Minput_focus_in can't trigger status_draw_cb, so call it here.
+    status_draw_cb (m_ic, Minput_status_draw);
 }
 
 void
 M17NInstance::focus_out ()
 {
     SCIM_DEBUG_IMENGINE(2) << "focus_out.\n";
-    minput_reset_ic(m_ic);
 
-    char buf [1024];
-    MText *produced;
-    int ret;
-    produced = mtext();
-    ret = minput_lookup(m_ic, Mnil, NULL, produced);
-
-    if (ret) {
-        SCIM_DEBUG_IMENGINE(3) << "minput_lookup returns 1\n";
-    }
-
-    mconv_rebind_buffer(__m17n_converter, (unsigned char *) buf, 1024);
-    mconv_encode(__m17n_converter, produced);
-    buf[__m17n_converter->nbytes] = 0;
-    m17n_object_unref(produced);
-
-    if (strlen (buf)) {
-        SCIM_DEBUG_IMENGINE(2) << "commit_string: " << buf << "\n";
-        commit_string (utf8_mbstowcs (buf));
-    }
+#if M17N_VERSION >= 10300
+    m17n_process_key (Minput_focus_out);
+#else
+    reset ();
+#endif
 }
 
 void
@@ -539,6 +462,12 @@ M17NInstance::trigger_property (const String &property)
 {
 }
 
+void
+M17NInstance::update_client_capabilities (unsigned int cap)
+{
+    m_cap = cap;
+}
+
 static M17NInstance *
 __find_instance (MInputContext *ic)
 {
@@ -556,17 +485,22 @@ MPlist *
 M17NInstance::register_callbacks(MPlist *callback_list)
 {
     if(!callback_list)
-      callback_list = mplist();
+        callback_list = mplist();
  
-    mplist_add(callback_list, Minput_preedit_start, (void*)preedit_start_cb);
-    mplist_add(callback_list, Minput_preedit_draw,  (void*)preedit_draw_cb);
-    mplist_add(callback_list, Minput_preedit_done,  (void*)preedit_done_cb);
-    mplist_add(callback_list, Minput_status_start,  (void*)status_start_cb);
-    mplist_add(callback_list, Minput_status_draw,   (void*)status_draw_cb);
-    mplist_add(callback_list, Minput_status_done,   (void*)status_done_cb);
-    mplist_add(callback_list, Minput_candidates_start, (void*)candidates_start_cb);
-    mplist_add(callback_list, Minput_candidates_draw,  (void*)candidates_draw_cb);
-    mplist_add(callback_list, Minput_candidates_done,  (void*)candidates_done_cb);
+    mplist_put(callback_list, Minput_preedit_start, (void*)preedit_start_cb);
+    mplist_put(callback_list, Minput_preedit_draw,  (void*)preedit_draw_cb);
+    mplist_put(callback_list, Minput_preedit_done,  (void*)preedit_done_cb);
+    mplist_put(callback_list, Minput_status_start,  (void*)status_start_cb);
+    mplist_put(callback_list, Minput_status_draw,   (void*)status_draw_cb);
+    mplist_put(callback_list, Minput_status_done,   (void*)status_done_cb);
+    mplist_put(callback_list, Minput_candidates_start, (void*)candidates_start_cb);
+    mplist_put(callback_list, Minput_candidates_draw,  (void*)candidates_draw_cb);
+    mplist_put(callback_list, Minput_candidates_done,  (void*)candidates_done_cb);
+
+#if M17N_VERSION >= 10300
+    mplist_put(callback_list, Minput_get_surrounding_text,  (void*)get_surrounding_text_cb);
+    mplist_put(callback_list, Minput_delete_surrounding_text,  (void*)delete_surrounding_text_cb);
+#endif
  
     return callback_list;
 }
@@ -576,9 +510,16 @@ M17NInstance::preedit_start_cb (MInputContext *ic, MSymbol command)
 {
     M17NInstance *this_ptr = __find_instance (ic);
 
-    if (this_ptr) {
+    if (this_ptr && !this_ptr->m_preedit_showed) {
         SCIM_DEBUG_IMENGINE(2) << "preedit_start_cb.\n";
+
+        if (this_ptr->m_block_preedit_op) {
+            this_ptr->m_pending_preedit_start = true;
+            return;
+        }
+
         this_ptr->show_preedit_string ();
+        this_ptr->m_preedit_showed = true;
     }
 }
 
@@ -590,6 +531,11 @@ M17NInstance::preedit_draw_cb  (MInputContext *ic, MSymbol command)
     if (this_ptr && ic->preedit) {
         SCIM_DEBUG_IMENGINE(2) << "preedit_draw_cb.\n";
 
+        if (this_ptr->m_block_preedit_op) {
+            this_ptr->m_pending_preedit_draw = true;
+            return;
+        }
+
         char buf[1024];
         mconv_rebind_buffer(__m17n_converter, (unsigned char *)buf, 1024);
         mconv_encode(__m17n_converter, ic->preedit);
@@ -605,11 +551,16 @@ M17NInstance::preedit_draw_cb  (MInputContext *ic, MSymbol command)
                                             SCIM_ATTR_DECORATE,
                                             SCIM_ATTR_DECORATE_REVERSE));
 
-            this_ptr->show_preedit_string ();
+            if (!this_ptr->m_preedit_showed) {
+                this_ptr->show_preedit_string ();
+                this_ptr->m_preedit_showed = true;
+            }
+
             this_ptr->update_preedit_string (wstr, attrs);
             this_ptr->update_preedit_caret (ic->cursor_pos);
         } else {
             this_ptr->hide_preedit_string ();
+            this_ptr->m_preedit_showed = false;
         }
     }
 }
@@ -619,9 +570,16 @@ M17NInstance::preedit_done_cb  (MInputContext *ic, MSymbol command)
 {
     M17NInstance *this_ptr = __find_instance (ic);
 
-    if (this_ptr) {
+    if (this_ptr && this_ptr->m_preedit_showed) {
         SCIM_DEBUG_IMENGINE(2) << "preedit_done_cb.\n";
+
+        if (this_ptr->m_block_preedit_op) {
+            this_ptr->m_pending_preedit_done = true;
+            return;
+        }
+
         this_ptr->hide_preedit_string ();
+        this_ptr->m_preedit_showed = false;
     }
 }
 
@@ -702,10 +660,12 @@ M17NInstance::candidates_draw_cb (MInputContext *ic, MSymbol command)
             MPlist *group;
             MText  *mt;
             char buf [1024];
-            int i, len, cur;
+            int i, len, cur, page;
+            bool page_up, page_down;
             WideString wstr;
 
             i = 0;
+            page = 0;
             group = ic->candidate_list;
             while (1) {
                 if (mplist_key (group) == Mtext)
@@ -717,10 +677,18 @@ M17NInstance::candidates_draw_cb (MInputContext *ic, MSymbol command)
                     break;
                 i += len;
                 group = mplist_next (group);
+
+                ++page;
             }
 
             cur = ic->candidate_index - i;
 
+            page_up = (page > 0);
+            page_down = ((page + 1) < mplist_length (ic->candidate_list));
+
+            if (page_up)
+                __lookup_table.append_candidate (0x3000);
+
             if (mplist_key (group) == Mtext) {
                 mt = (MText *) mplist_value (group);
                 mconv_rebind_buffer(__m17n_converter, (unsigned char *)buf, 1024);
@@ -731,6 +699,11 @@ M17NInstance::candidates_draw_cb (MInputContext *ic, MSymbol command)
                 for (i = 0; i < wstr.length (); ++i)
                     __lookup_table.append_candidate (wstr [i]);
 
+                if (page_up) {
+                    __lookup_table.set_page_size (1);
+                    __lookup_table.page_down ();
+                }
+
                 __lookup_table.set_page_size (wstr.length ());
             } else {
                 MPlist *pl;
@@ -746,9 +719,18 @@ M17NInstance::candidates_draw_cb (MInputContext *ic, MSymbol command)
                     __lookup_table.append_candidate (wstr);
                     ++ len;
                 }
+
+                if (page_up) {
+                    __lookup_table.set_page_size (1);
+                    __lookup_table.page_down ();
+                }
+
                 __lookup_table.set_page_size (len);
             }
 
+            if (page_down)
+                __lookup_table.append_candidate (0x3000);
+
             __lookup_table.set_cursor_pos_in_current_page (cur);
             __lookup_table.show_cursor ();
 
@@ -771,6 +753,45 @@ M17NInstance::candidates_done_cb (MInputContext *ic, MSymbol command)
     }
 }
 
+#if M17N_VERSION >= 10300
+void
+M17NInstance::get_surrounding_text_cb (MInputContext *ic, MSymbol command)
+{
+    M17NInstance *this_ptr = __find_instance (ic);
+
+    if (this_ptr && (this_ptr->m_cap & SCIM_CLIENT_CAP_SURROUNDING_TEXT)) {
+        SCIM_DEBUG_IMENGINE(2) << "get_surrounding_text_cb.\n";
+        if (ic->plist && mplist_key (ic->plist) == Minteger) {
+            int len = (int) ((size_t) mplist_value (ic->plist));
+            int cursor;
+            WideString wstr;
+            MText *txt = mtext ();
+            if (this_ptr->get_surrounding_text (wstr, cursor, (len < 0) ? (-len) : 0, (len > 0) ? len : 0)) {
+                for (WideString::const_iterator i = wstr.begin (); i != wstr.end (); ++i)
+                    mtext_cat_char (txt, (int) *i);
+            }
+            mplist_set (ic->plist, Mtext, txt);
+            m17n_object_unref (txt);
+        }
+    }
+}
+
+void
+M17NInstance::delete_surrounding_text_cb (MInputContext *ic, MSymbol command)
+{
+    M17NInstance *this_ptr = __find_instance (ic);
+
+    if (this_ptr) {
+        SCIM_DEBUG_IMENGINE(2) << "delete_surrounding_text_cb.\n";
+        if (ic->plist && mplist_key (ic->plist) == Minteger) {
+            int len = (int) ((size_t) mplist_value (ic->plist));
+            // FIXME: There is no way to inform m17n whether it's failed or not.
+            this_ptr->delete_surrounding_text ((len < 0) ? len : 0, (len > 0) ? len : (-len));
+        }
+    }
+}
+#endif
+
 static MSymbol
 __key_to_symbol (const KeyEvent &key)
 {
@@ -802,6 +823,9 @@ __key_to_symbol (const KeyEvent &key)
             mask |= SCIM_KEY_ShiftMask;
     }
 
+    if (key.is_meta_down ())
+        mask |= SCIM_KEY_MetaMask;
+
     if (key.is_alt_down ())
         mask |= SCIM_KEY_AltMask;
 
@@ -811,8 +835,10 @@ __key_to_symbol (const KeyEvent &key)
         keysym = String ("S-") + keysym;
     if (mask & SCIM_KEY_ControlMask)
         keysym = String ("C-") + keysym;
-    if (mask & SCIM_KEY_AltMask)
+    if (mask & SCIM_KEY_MetaMask)
         keysym = String ("M-") + keysym;
+    if (mask & SCIM_KEY_AltMask)
+        keysym = String ("A-") + keysym;
 
     return msymbol (keysym.c_str ());
 }
